from django_celery_beat.models import PeriodicTask, CrontabSchedule
from django_celery_results.models import TaskResult
from rest_framework import serializers


from utg_base.celery.serializers import TaskResultSerializer


class CrontabScheduleSerializer(serializers.ModelSerializer):
    timezone = serializers.SerializerMethodField()

    class Meta:
        model = CrontabSchedule
        fields = '__all__'

    def get_timezone(self, obj):
        return str(obj.timezone) if obj.timezone else None


class PeriodicTaskSerializer(serializers.ModelSerializer):
    crontab = CrontabScheduleSerializer(read_only=True)
    task_result = serializers.SerializerMethodField(read_only=True)

    class Meta:
        model = PeriodicTask
        fields = '__all__'
        read_only_fields = ['name', 'task', 'args', 'kwargs', 'queue', 'exchange', 'routing_key', 'headers', 'priority',
                            'expires', 'expire_seconds', 'one_off', 'start_time', 'interval', 'solar', 'clocked']

    def get_task_result(self, obj):
        latest_task_result = TaskResult.objects.filter(
            periodic_task_name=obj.name
        ).order_by('-date_created').first()

        if not latest_task_result:
            return None

        return TaskResultSerializer(latest_task_result).data



import traceback
from typing import Literal

from django.http import HttpRequest
from django.urls import URLResolver, URLPattern, get_resolver
from django_redis import get_redis_connection
from redis.client import Redis
from rest_framework.request import Request

from utg_base.env import env
from utg_base.u_services import u_requests
from utg_base.u_services.constants import UServices
from utg_base.utils import to_snake_case
from utg_base.utils.data import safe_get


def generate_perm_cache_key(user_id: str):
    return f"user:{user_id}:permissions"


def has_perm(user_id, perm):
    redis_conn: Redis = get_redis_connection("shared")
    perm = f"{to_snake_case(env('APP_NAME'))}:{perm}"
    return bool(redis_conn.sismember(generate_perm_cache_key(user_id), perm))


def has_perms(user_id: str, perms: list[str], operator: Literal["OR", "AND"] = "OR"):
    redis_conn: Redis = get_redis_connection("shared")
    perms = [f"{to_snake_case(env('APP_NAME'))}:{perm}" for perm in perms]
    result = redis_conn.smismember(generate_perm_cache_key(user_id), perms)

    if operator == "OR":
        return any(result)
    return all(result)


def get_permissions(url_patterns):
    permissions = set()

    for item in url_patterns:
        if isinstance(item, URLPattern):
            view = item.callback

            view_cls = safe_get(view, 'view_class')
            # for View
            if view_cls:
                for method_name in ["get", "post", "put", "patch", "delete"]:
                    perms = safe_get(view_cls, f'{method_name}._perms', set())
                    permissions.update(perms)
            # for ViewSet
            else:
                view_cls = safe_get(view, 'cls')
                actions: dict = safe_get(view, 'actions', {})
                for method_name in actions.values():
                    perms = safe_get(view_cls, f'{method_name}._perms', set())
                    permissions.update(perms)

        elif isinstance(item, URLResolver):
            permissions.update(get_permissions(
                item.url_patterns
            ))

    return permissions


def sync_permissions():
    resolver = get_resolver()
    perm_codes = list(get_permissions(resolver.url_patterns))

    request = Request(HttpRequest())
    try:
        response = u_requests.post(request, UServices.USER_MANAGEMENT, '/api/permissions/', data={
            "service": to_snake_case(env('APP_NAME')),
            "codes": perm_codes
        })
        print('[INFO] sync_permissions:', response.json())
    except Exception as e:
        traceback.print_exc()



import re

def to_snake_case(s: str) -> str:
    """
    Convert any string into snake_case.
    Handles:
      - camelCase / PascalCase
      - mixed separators (-, space, ., !, etc.)
      - multiple underscores
      - trims leading/trailing separators
    """
    if not isinstance(s, str):
        raise TypeError("Input must be a string")

    # Trim spaces
    s = s.strip()
    if not s:
        return ""

    # Insert underscore between lower/number and upper case letters
    # "helloWorld" -> "hello_World"
    s = re.sub(r'(?<=[0-9a-z])(?=[A-Z])', '_', s)

    # Insert underscore between acronyms and normal words
    # "HTTPResponse" -> "HTTP_Response"
    s = re.sub(r'(?<=[A-Z])(?=[A-Z][a-z])', '_', s)

    # Replace all non-word chars with underscore
    # \W = anything except letters, digits, underscore
    s = re.sub(r'\W+', '_', s)

    # Replace spaces if any remain
    s = s.replace(" ", "_")

    # Collapse multiple underscores
    s = re.sub(r'_+', '_', s)

    # Remove leading/trailing underscores
    s = s.strip('_')

    return s.lower()



import importlib
import json

from django.core.management.base import BaseCommand
from django_celery_beat.models import CrontabSchedule, PeriodicTask


class Command(BaseCommand):
    help = "Migrate tasks"

    def add_arguments(self, parser):
        parser.add_argument('--app', type=str, help="An optional argument", required=True)

    def handle(self, *args, **options):
        try:
            tasks = importlib.import_module(f"{options['app']}.tasks")
            tasks = tasks.tasks
            for task in tasks:
                celery_task, _ = PeriodicTask.objects.update_or_create(
                    name=task['name'],
                    defaults={
                        'crontab': self.get_crontab_schedule(task['crontab']),
                        'task': task['task'].name,
                        'args': json.dumps(task.get('args') or []),
                        'kwargs': json.dumps(task.get('kwargs') or {}),
                        'enabled': task.get('enabled', True),
                    }
                )
            self.stdout.write(
                self.style.SUCCESS('Successfully migrated tasks')
            )
        except ModuleNotFoundError:
            self.stdout.write(self.style.ERROR("Tasks module not found"))

    @staticmethod
    def get_crontab_schedule(crontab='* * * * *'):
        minute, hour, day_of_month, month_of_year, day_of_week = crontab.split(' ')
        cron, _ = CrontabSchedule.objects.get_or_create(
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month_of_year=month_of_year,
            day_of_week=day_of_week,
        )
        return cron



import multiprocessing
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Callable, Any, Iterable, Tuple, Dict, Union


class ThreadPoolException(Exception):
    """Exception to wrap the error in each task."""

    def __init__(self, func: str, exception: Exception):
        self.func = func
        self.exception = exception
        super().__init__(f"Exception in {func}: {exception}")


def parallel_execute(
        *tasks: Union[
            Callable,  # func
            Tuple[Callable, Iterable],  # (func, args)
            Tuple[Callable, Iterable, Dict[str, Any]]  # (func, args, kwargs)
        ],
        max_workers: int = None
) -> list[Any]:
    """
    Executes multiple functions in parallel.
    Each can be without arguments or with args/kwargs.

    Example:
        parallel_execute(func1, func2)
        parallel_execute((func1, (1, 2)), (func2, (), {"x": 5}))

    The sequence is preserved.

    :return: Result list or ThreadPoolException objects
    """
    if not tasks:
        return []

    cpu_count = multiprocessing.cpu_count()
    max_workers = max_workers or min(cpu_count * 10, len(tasks))
    results = [None] * len(tasks)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        futures = {}

        for i, task in enumerate(tasks):
            # Task unpacking
            if callable(task):
                func, args, kwargs = task, (), {}
            elif isinstance(task, tuple):
                func = task[0]
                args = task[1] if len(task) > 1 else ()
                kwargs = task[2] if len(task) > 2 else {}
            else:
                raise ValueError(f"Invalid task format: {task}")

            futures[executor.submit(func, *args, **kwargs)] = i

        for future in as_completed(futures):
            idx = futures[future]
            func = tasks[idx][0] if isinstance(tasks[idx], tuple) else tasks[idx]
            func_name = getattr(func, "__name__", str(func))
            try:
                results[idx] = future.result()
            except Exception as e:
                results[idx] = ThreadPoolException(func=func_name, exception=e)

    return results



import django_filters
from django_celery_results.models import TaskResult
from drf_spectacular.utils import extend_schema
from rest_framework import viewsets, filters

from utg_base.celery.filters import TaskResultFilterSet
from utg_base.api.permissions import IsSuperUser
from utg_base.celery.serializers import TaskResultSerializer


@extend_schema(tags=['admin/task-results'])
class TaskResultViewSet(viewsets.ReadOnlyModelViewSet):
    queryset = TaskResult.objects.all()
    serializer_class = TaskResultSerializer
    permission_classes = [IsSuperUser]
    filterset_class = TaskResultFilterSet
    filter_backends = [filters.SearchFilter, django_filters.rest_framework.DjangoFilterBackend]
    search_fields = ['name']



from django.conf import settings
from django.http import HttpResponseForbidden
from django.urls import path
from drf_spectacular.views import SpectacularAPIView, SpectacularSwaggerView
from rest_framework.views import APIView

from utg_base.env import env


class DebugTokenRequiredMixin(APIView):
    """
    If Request header X-Debug-Token not found then 403
    """

    def dispatch(self, request, *args, **kwargs):
        token = request.headers.get("X-Debug-Token")
        if not settings.DEBUG and token != env("DJANGO_SECRET_KEY"):
            return HttpResponseForbidden("Forbidden")
        return super().dispatch(request, *args, **kwargs)


class ProtectedSpectacularAPIView(DebugTokenRequiredMixin, SpectacularAPIView):
    pass


class ProtectedSpectacularSwaggerView(DebugTokenRequiredMixin, SpectacularSwaggerView):
    pass


urlpatterns = [
    path('api/swagger/schema/', ProtectedSpectacularAPIView.as_view(), name='schema'),
    path('api/docs/', ProtectedSpectacularSwaggerView.as_view(url_name='schema'), name='docs'),
]



from django.urls import path

from utg_base.api.routers import OptionalSlashRouter
from utg_base.celery.views import TaskResultViewSet, PeriodicTaskViewSet, PeriodicTaskRunNowView

router = OptionalSlashRouter()
router.register('task-results', TaskResultViewSet, basename='task-results')
router.register('periodic-tasks', PeriodicTaskViewSet, basename='periodic-tasks')

urlpatterns = [
    path('periodic-tasks/<pk>/run-now/', PeriodicTaskRunNowView.as_view(), name='periodic-tasks-run-now'),
]

urlpatterns += router.urls



from django.middleware.locale import LocaleMiddleware as DjangoLocaleMiddleware
from django.utils import translation

from utg_base.constants import AVAILABLE_LANGUAGES


class LocaleMiddleware(DjangoLocaleMiddleware):
    def process_request(self, request):
        if lang_from_header := request.headers.get('accept-language'):
            if lang_from_header and lang_from_header not in AVAILABLE_LANGUAGES:
                lang_from_header = 'ru'
        translation.activate(lang_from_header)
        request.LANGUAGE_CODE = translation.get_language()



from django.utils.deprecation import MiddlewareMixin

from utg_base.env import env


class DebugOverrideMiddleware(MiddlewareMixin):
    def process_exception(self, request, exception):
        if request.headers.get("X-Debug-Token") == env("DJANGO_SECRET_KEY"):
            from django.views import debug
            return debug.technical_500_response(request, type(exception), exception, exception.__traceback__)



from functools import wraps
from typing import Literal, Dict, Union, List

from rest_framework.exceptions import PermissionDenied

from .utils import has_perms as has_permissions


def has_perm(*permission: str, operator: Literal["OR", "AND"] = "OR"):
    """
    Usage:
        has_perm('PERM1')
        has_perm('PERM1', 'PERM2')

    :param permission: One or many permissions.
    """
    def decorator(view_func):
        view_func._perms = permission
        @wraps(view_func)
        def _wrapped_view(self, request, *args, **kwargs):
            if not has_permissions(user_id=request.user.id, perms=list(permission), operator=operator):
                raise PermissionDenied
            return view_func(self, request, *args, **kwargs)
        return _wrapped_view
    return decorator


def has_class_perm(permissions_map: Dict[str, Union[str, List[str]]], operator: Literal["OR", "AND"] = "OR"):
    """
    Usage:
        @has_class_perm({
            'list': ['perm1', 'perm2'],
            'create': 'perm3',
            'update': ['perm4']
        })
    """
    def decorator(cls):
        for action, perms in permissions_map.items():
            if isinstance(perms, str):
                perms = [perms]

            if hasattr(cls, action):
                method = getattr(cls, action)
                setattr(method, "_perms", perms)
        original_initial = cls.initial

        @wraps(original_initial)
        def new_initial(self, request, *args, **kwargs):
            action = getattr(self, 'action', None) or request.method.lower()

            if action in permissions_map:
                required_perms = permissions_map[action]

                if isinstance(required_perms, str):
                    required_perms = [required_perms]

                if not has_permissions(user_id=request.user.id, perms=required_perms, operator=operator):
                    raise PermissionDenied

            return original_initial(self, request, *args, **kwargs)

        cls.initial = new_initial
        return cls

    return decorator



import requests
from rest_framework.request import Request

from .constants import UServices
from .models import UService


def u_request(
        method: str,
        request: Request,
        u_service: UServices,
        prefix: str,
        params: dict = None,
        data: dict | list = None
):
    headers = {
        'Authorization': request.headers.get('Authorization'),
        'Accept-Language': request.headers.get('Accept-Language'),
    }
    return requests.request(
        method=method,
        url=UService.get_by_name(u_service.value).get_url(prefix),
        params=params,
        json=data,
        headers=headers,
    )


def get(request: Request, u_service: UServices, prefix: str, params: dict = None):
    return u_request(method='get', request=request, u_service=u_service, prefix=prefix, params=params)


def post(request: Request, u_service: UServices, prefix: str, params: dict = None, data: dict | list = None):
    return u_request(method='post', request=request, u_service=u_service, prefix=prefix, params=params, data=data)


def put(request: Request, u_service: UServices, prefix: str, params: dict = None, data: dict | list = None):
    return u_request(method='put', request=request, u_service=u_service, prefix=prefix, params=params, data=data)


def patch(request: Request, u_service: UServices, prefix: str, params: dict = None, data: dict | list = None):
    return u_request(method='patch', request=request, u_service=u_service, prefix=prefix, params=params, data=data)


def delete(request: Request, u_service: UServices, prefix: str, params: dict = None):
    return u_request(method='delete', request=request, u_service=u_service, prefix=prefix, params=params)



import os

import dotenv
import hvac
from django.conf import settings

dotenv.load_dotenv(settings.BASE_DIR / '.env')

assert os.environ.get('VAULT_URL') is not None, "VAULT_URL not defined, please set VAULT_URL on .env file"
assert os.environ.get('VAULT_TOKEN') is not None, "VAULT_TOKEN not defined, please set VAULT_TOKEN on .env file"
assert os.environ.get('VAULT_PATH') is not None, "VAULT_PATH not defined, please set VAULT_PATH on .env file"

client = hvac.Client(
    url=os.environ.get("VAULT_URL"),
    token=os.environ.get("VAULT_TOKEN"),
    verify=False
)
if not client.is_authenticated():
    raise Exception("Vault authentication failed")

envs = client.secrets.kv.read_secret_version(
    path=os.environ.get("VAULT_PATH"),
    mount_point="utg-scada"
)['data']['data']


def env(key, default=None):
    if key in os.environ:
        return os.environ.get(key, default)
    else:
        return envs.get(key, default)
